# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
from .....schema_classes import CalendarIntervalClass
from .....schema_classes import PartitionSpecClass
from .....schema_classes import PartitionTypeClass
from .....schema_classes import TimeWindowClass
from .....schema_classes import TimeWindowSizeClass


CalendarInterval = CalendarIntervalClass
PartitionSpec = PartitionSpecClass
PartitionType = PartitionTypeClass
TimeWindow = TimeWindowClass
TimeWindowSize = TimeWindowSizeClass

# fmt: on
